from unittest.mock import patch

import paho.mqtt.client as mqtt
from bx_py_utils.test_utils.redirect import RedirectOut
from bx_py_utils.test_utils.snapshot import assert_snapshot
from cli_base.cli_tools.test_utils.rich_test_utils import NoColorEnvRichClick
from manageprojects.tests.base import BaseTestCase

from pysmartmeter import __version__, publish_loop
from pysmartmeter.data_classes import MqttSettings
from pysmartmeter.publish_loop import logger as publish_forever_logger
from pysmartmeter.publish_loop import publish_forever
from pysmartmeter.tests.data import TEST_DATA_BIG
from pysmartmeter.tests.mocks import MqttClientMock, SerialMock, SerialMockEnds, SocketMock


class CliTestCase(BaseTestCase):
    def test_publish_loop(self):
        mocked_serial = SerialMock(lines=TEST_DATA_BIG)
        mqtt_client = MqttClientMock()
        socket_mock = SocketMock()

        with (
            patch.object(publish_loop, 'get_serial', mocked_serial),
            patch.object(mqtt, 'Client', mqtt_client),
            patch.object(publish_loop, 'socket', socket_mock),
            self.assertRaises(SerialMockEnds),
            NoColorEnvRichClick(),
            RedirectOut() as buffer,
        ):
            publish_forever(
                settings=MqttSettings(
                    host='foo.host.tld',
                    port=123,
                    user_name='bar',
                    password='foobarbaz',
                ),
                verbose=True,
            )

        mqtt_client.assert_state(
            self,
            init=[[(), {'client_id': f'PySmartMeter v{__version__} on mocked get hostname'}]],
            enabled_logger=[publish_forever_logger],
            credentials=[{'password': 'foobarbaz', 'user_name': 'bar'}],
            connects=[{'host': 'foo.host.tld', 'port': 123}],
            loop_started=True,
        )
        published = mqtt_client.published
        self.assertGreaterEqual(len(published), 8)

        self.assertEqual(buffer.stderr, '')
        self.assert_in_content(
            got=buffer.stdout,
            parts=(
                f'PySmartMeter v{__version__}',
                'Connect foo.host.tld:123 ',
                'login with user: bxr password:fxxxxxxxz... OK',
                'Publish MQTT topic: homeassistant/sensor/EBZ5DD3BZ06ETA107/state',
            ),
        )

        self.assertEqual(socket_mock.getaddrinfos, [('foo.host.tld', 123)])

        assert_snapshot(got=published)



class SerialMockEnds(StopIteration):
    pass


class SerialMock:
    def __init__(self, lines):
        self.lines = iter(lines)

    def __call__(self):
        return self

    def readline(self):
        try:
            return next(self.lines)
        except StopIteration:
            raise SerialMockEnds


class MqttClientMock:
    def __init__(self):
        self.init = []
        self.enabled_logger = []
        self.credentials = []
        self.connects = []
        self.loop_started = False
        self.published = []

    def __call__(self, *args, **kwargs):
        self.init.append([args, kwargs])
        return self

    def enable_logger(self, logger):
        self.enabled_logger.append(logger)

    def username_pw_set(self, user_name, password):
        self.credentials.append(dict(user_name=user_name, password=password))

    def connect(self, host, port):
        self.connects.append(dict(host=host, port=port))

    def loop_start(self):
        self.loop_started = True

    def is_connected(self):
        if self.connects and self.loop_started:
            return True
        return False

    def publish(self, **kwargs):
        self.published.append(kwargs)

    def assert_state(self, test_class, *, init, enabled_logger, credentials, connects, loop_started):
        test_class.assertEqual(self.init, init)
        test_class.assertEqual(self.enabled_logger, enabled_logger)
        test_class.assertEqual(self.credentials, credentials)
        test_class.assertEqual(self.connects, connects)
        test_class.assertEqual(self.loop_started, loop_started)

    def loop_stop(self):
        pass

    def disconnect(self):
        pass


class MqttPublisherMock:
    mqtt_payloads = []

    def publish(self, *, mqtt_payload):
        self.mqtt_payloads.append(mqtt_payload)


class SocketMock:
    def __init__(self):
        self.getaddrinfos = []

    def gethostname(self):
        return 'mocked get hostname'

    def setdefaulttimeout(self, timeout):
        pass

    def getaddrinfo(self, host, port):
        self.getaddrinfos.append((host, port))
        return True



import json
import logging
import socket
import sys

import paho.mqtt.client as mqtt
from bx_py_utils.anonymize import anonymize
from rich import print
from rich.pretty import pprint

import pysmartmeter
from pysmartmeter.data_classes import MqttPayload, MqttSettings, ObisValue
from pysmartmeter.detect_serial import get_serial
from pysmartmeter.homeassistant import data2config, data2state, ha_convert_obis_values
from pysmartmeter.parser import ObisParser


logger = logging.getLogger(__name__)


def get_client_id():
    hostname = socket.gethostname()
    client_id = f'PySmartMeter v{pysmartmeter.__version__} on {hostname}'
    return client_id


class OnConnectCallback:
    def __init__(self, verbose: bool = True):
        self.verbose = verbose

    def __call__(self, client, userdata, flags, rc):
        if self.verbose:
            print(f'MQTT broker connect result code: {rc}', end=' ')

        if rc == 0:
            if self.verbose:
                print('[green]OK')
        else:
            print('\n[red]MQTT Connection not successful!')
            print('[yellow]Please check your credentials\n')
            raise RuntimeError(f'MQTT connection result code {rc} is not 0')

        if self.verbose:
            print(f'\t{userdata=}')
            print(f'\t{flags=}')


def get_connected_client(settings: MqttSettings, verbose: bool = True, timeout=10):
    client_id = get_client_id()

    if verbose:
        print(
            f'\nConnect [cyan]{settings.host}:{settings.port}[/cyan] as "[magenta]{client_id}[/magenta]"...', end=' '
        )

    socket.setdefaulttimeout(timeout)  # Sadly: Timeout will not used in getaddrinfo()!
    info = socket.getaddrinfo(settings.host, settings.port)
    if not info:
        print('[red]Resolve error: No info!')
    elif verbose:
        print('Host/port test [green]OK')

    mqttc = mqtt.Client(client_id=client_id)
    mqttc.on_connect = OnConnectCallback(verbose=verbose)
    mqttc.enable_logger(logger=logger)

    if settings.user_name and settings.password:
        if verbose:
            print(
                f'login with user: {anonymize(settings.user_name)} password:{anonymize(settings.password)}...',
                end=' ',
            )
        mqttc.username_pw_set(settings.user_name, settings.password)

    mqttc.connect(settings.host, port=settings.port)

    if verbose:
        print('[green]OK')
    return mqttc


class MqttPublisher:
    def __init__(self, settings: MqttSettings, verbose: bool = True):
        self.verbose = verbose
        self.mqttc = get_connected_client(settings=settings, verbose=verbose)
        self.mqttc.loop_start()

    def publish(self, *, mqtt_payload: MqttPayload) -> None:
        if self.verbose:
            print('_' * 100)
            print(f'Publish MQTT topic: {mqtt_payload.topic}')
            pprint(mqtt_payload.data)

        assert self.mqttc.is_connected(), 'Not connected to MQTT broker!'
        info = self.mqttc.publish(topic=mqtt_payload.topic, payload=json.dumps(mqtt_payload.data))

        if self.verbose:
            print('publish result:', info)


def obis_values2mqtt_state(obis_values: list[ObisValue], verbose: bool = False) -> MqttPayload:
    ha_values = ha_convert_obis_values(obis_values=obis_values)
    if verbose:
        pprint(ha_values)
    return data2state(ha_values=ha_values)


def obis_values2mqtt_config(obis_values: list[ObisValue]) -> list[MqttPayload]:
    ha_values = ha_convert_obis_values(obis_values=obis_values)
    return data2config(ha_values=ha_values)


class HomeAssistantMqtt:
    def __init__(self, mqtt_publisher: MqttPublisher, verbose: bool = True):
        self.mqtt_publisher = mqtt_publisher
        self.verbose = verbose

        self.send_count = 0

    def __call__(self, *, obis_values: list[ObisValue]):
        if self.send_count % 12 == 0:
            for payload in obis_values2mqtt_config(obis_values=obis_values):
                self.mqtt_publisher.publish(mqtt_payload=payload)

        if self.verbose:
            print(f'publish values (send count: {self.send_count}):')
        self.mqtt_publisher.publish(
            mqtt_payload=obis_values2mqtt_state(obis_values=obis_values, verbose=self.verbose)
        )
        self.send_count += 1


def publish_forever(*, settings: MqttSettings, verbose):
    serial = get_serial()
    if not serial:
        print('Serial not found')
        sys.exit(1)

    mqtt_publisher = MqttPublisher(settings=settings, verbose=verbose)
    publisher = HomeAssistantMqtt(mqtt_publisher=mqtt_publisher, verbose=verbose)
    parser = ObisParser(publish_callback=publisher, verbose=verbose)

    while True:
        data = serial.readline()
        parser.feed_line(data)



from unittest.mock import patch

import paho.mqtt.client as mqtt
from bx_py_utils.test_utils.snapshot import assert_text_snapshot
from manageprojects.test_utils.click_cli_utils import invoke_click
from manageprojects.tests.base import BaseTestCase

import pysmartmeter
from pysmartmeter import publish_loop
from pysmartmeter.cli import cli_app
from pysmartmeter.cli.cli_app import cli
from pysmartmeter.data_classes import MqttSettings
from pysmartmeter.tests.mocks import MqttClientMock, SocketMock


class TestMqttConnection(BaseTestCase):
    def test_via_cli(self):
        mqtt_client = MqttClientMock()
        socket_mock = SocketMock()
        settings = MqttSettings(
            host='foo.host.tld',
            port=123,
            user_name='bar',
            password='foobarbaz',
        )

        with patch.object(mqtt, 'Client', mqtt_client), patch.object(
            publish_loop, 'socket', socket_mock
        ), patch.object(cli_app, 'get_mqtt_settings', return_value=settings), patch.object(
            pysmartmeter, '__version__', '1.2.3'
        ):
            stdout = invoke_click(cli, 'test-mqtt-connection')

        self.assert_in_content(
            got=stdout,
            parts=(
                'Connect foo.host.tld:123 ',
                'Host/port test OK',
                'Test succeed',
            ),
        )
        assert_text_snapshot(got=stdout)



from bx_py_utils.test_utils.unittest_utils import BaseDocTests

import pysmartmeter


class DocTests(BaseDocTests):
    def test_doctests(self):
        self.run_doctests(
            modules=(pysmartmeter,),
        )



from unittest.mock import patch

from manageprojects.test_utils.click_cli_utils import invoke_click
from manageprojects.tests.base import BaseTestCase

from pysmartmeter import detect_serial
from pysmartmeter.cli import cli_app
from pysmartmeter.cli.cli_app import cli
from pysmartmeter.data_classes import MqttSettings


class CliTestCase(BaseTestCase):
    def test_publish_loop(self):
        settings = MqttSettings(
            host='foo.host.tld',
            port=123,
            user_name='bar',
            password='foobarbaz',
        )

        class PublishForever:
            calls = []

            def __call__(self, **kwargs):
                self.calls.append(kwargs)
                raise KeyboardInterrupt

        mocked_publish_forever = PublishForever()

        with patch.object(cli_app, 'get_mqtt_settings', return_value=settings), patch.object(
            cli_app, 'publish_forever', mocked_publish_forever
        ):
            stdout = invoke_click(cli, 'publish-loop')

        self.assertEqual(mocked_publish_forever.calls, [{'settings': settings, 'verbose': True}])
        self.assert_in_content(
            got=stdout,
            parts=(
                'foo.host.tld',
                'fxxxxxxxz',
                'Bye, bye',
            ),
        )

    def test_detect_serial(self):
        with patch.object(detect_serial, 'comports', return_value=[]):
            stdout = invoke_click(cli, 'detect-serial')

        self.assert_in_content(
            got=stdout,
            parts=(
                'Detect Serial...',
                'No serial ports found!',
            ),
        )



"""
    CLI for usage
"""
import getpass
import logging
import sys
from pathlib import Path

import rich_click as click
from cli_base.cli_tools.version_info import print_version
from rich import print  # noqa
from rich.console import Console
from rich.pretty import pprint
from rich.traceback import install as rich_traceback_install
from rich_click import RichGroup

import pysmartmeter
from pysmartmeter import constants
from pysmartmeter.data_classes import MqttSettings
from pysmartmeter.detect_serial import print_detect_serial
from pysmartmeter.dump import serial_dump
from pysmartmeter.log_utils import log_config
from pysmartmeter.publish_loop import get_connected_client, publish_forever
from pysmartmeter.utilities import systemd
from pysmartmeter.utilities.credentials import get_mqtt_settings, store_mqtt_settings


logger = logging.getLogger(__name__)


OPTION_ARGS_DEFAULT_TRUE = dict(is_flag=True, show_default=True, default=True)
OPTION_ARGS_DEFAULT_FALSE = dict(is_flag=True, show_default=True, default=False)
ARGUMENT_EXISTING_DIR = dict(
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=Path)
)
ARGUMENT_NOT_EXISTING_DIR = dict(
    type=click.Path(exists=False, file_okay=False, dir_okay=True, readable=False, writable=True, path_type=Path)
)
ARGUMENT_EXISTING_FILE = dict(
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path)
)


class ClickGroup(RichGroup):  # FIXME: How to set the "info_name" easier?
    def make_context(self, info_name, *args, **kwargs):
        info_name = './cli.py'
        return super().make_context(info_name, *args, **kwargs)


@click.group(
    cls=ClickGroup,
    epilog=constants.CLI_EPILOG,
)
def cli():
    pass


@click.command()
def detect_serial():
    """
    Just print the detected serial port instance
    """
    log_config()
    print_detect_serial()


cli.add_command(detect_serial)


@click.command()
def dump():
    """
    Just dump serial output
    """
    log_config()
    serial_dump()


cli.add_command(dump)


@click.command()
@click.option('--log/--no-log', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_TRUE)
def publish_loop(log: bool = True, verbose: bool = True):
    """
    Publish current data via MQTT (endless loop)
    """
    if log:
        log_config()
    settings: MqttSettings = get_mqtt_settings()
    pprint(settings.anonymized())
    try:
        publish_forever(settings=settings, verbose=verbose)
    except KeyboardInterrupt:
        print('Bye, bye')


cli.add_command(publish_loop)


@click.command()
def store_settings():
    """
    Store MQTT server settings.
    """
    log_config()

    try:
        settings: MqttSettings = get_mqtt_settings()
    except FileNotFoundError:
        print('No settings stored, yet. ok.')
        print()
        print('Input settings:')
    else:
        print('Current settings:')
        pprint(settings.anonymized())
        print()
        print('Input new settings:')

    host = input('host (e.g.: "test.mosquitto.org"): ')
    if not host:
        print('Host is needed! Abort.')
        sys.exit(1)

    port = input('port (default: 1883): ')
    if port:
        port = int(port)
    else:
        port = 1883
    user_name = input('user name: ')
    password = getpass.getpass('password: ')

    settings = MqttSettings(host=host, port=port, user_name=user_name, password=password)
    file_path = store_mqtt_settings(settings)
    print(f'MQTT server settings stored into: {file_path}')


cli.add_command(store_settings)


@click.command()
def debug_settings():
    """
    Display (anonymized) MQTT server username and password
    """
    log_config()
    settings: MqttSettings = get_mqtt_settings()
    pprint(settings.anonymized())


cli.add_command(debug_settings)


@click.command()
def test_mqtt_connection():
    """
    Test connection to MQTT Server
    """
    log_config()
    settings: MqttSettings = get_mqtt_settings()
    mqttc = get_connected_client(settings=settings, verbose=True)
    mqttc.loop_start()
    mqttc.loop_stop()
    mqttc.disconnect()
    print('\n[green]Test succeed[/green], bye ;)')


cli.add_command(test_mqtt_connection)


@click.command()
def debug_systemd_service():
    """
    Just print the systemd service file content
    """
    content = systemd.compile_service()
    print(content)


cli.add_command(debug_systemd_service)


@click.command()
def setup_systemd_service():
    """
    Setup PySmartMeter systemd services and starts it.
    """
    systemd.write_service_file()
    systemd.enable_service()
    systemd.restart_service()
    systemd.status()


cli.add_command(setup_systemd_service)


@click.command()
def systemd_status():
    """
    Call systemd status of PySmartMeter services
    """
    systemd.status()


cli.add_command(systemd_status)


@click.command()
def systemd_stop():
    """
    Stop PySmartMeter systemd services
    """
    systemd.stop()
    systemd.status()


cli.add_command(systemd_stop)


@click.command()
def systemd_restart():
    """
    Restart PySmartMeter systemd services
    """
    systemd.restart_service()
    systemd.status()


cli.add_command(systemd_restart)


@click.command()
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


cli.add_command(version)


def main():
    print_version(pysmartmeter)

    console = Console()
    rich_traceback_install(
        width=console.size.width,  # full terminal width
        show_locals=True,
        suppress=[click],
        max_frames=2,
    )

    # Execute Click CLI:
    cli.name = './cli.py'
    cli()



"""
    CLI for development
"""
import logging
import sys
from pathlib import Path

import rich_click as click
from bx_py_utils.path import assert_is_file
from cli_base.cli_tools import code_style
from cli_base.cli_tools.dev_tools import run_coverage, run_tox, run_unittest_cli
from cli_base.cli_tools.subprocess_utils import verbose_check_call
from cli_base.cli_tools.test_utils.snapshot import UpdateTestSnapshotFiles
from cli_base.cli_tools.verbosity import OPTION_KWARGS_VERBOSE
from cli_base.cli_tools.version_info import print_version
from manageprojects.utilities.publish import publish_package
from rich.console import Console
from rich.traceback import install as rich_traceback_install
from rich_click import RichGroup

import pysmartmeter
from pysmartmeter import constants
from pysmartmeter.constants import BASE_PATH


logger = logging.getLogger(__name__)


PACKAGE_ROOT = BASE_PATH.parent
assert_is_file(PACKAGE_ROOT / 'pyproject.toml')  # Exists only in cloned git repo


OPTION_ARGS_DEFAULT_TRUE = dict(is_flag=True, show_default=True, default=True)
OPTION_ARGS_DEFAULT_FALSE = dict(is_flag=True, show_default=True, default=False)
ARGUMENT_EXISTING_DIR = dict(
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=Path)
)
ARGUMENT_NOT_EXISTING_DIR = dict(
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        readable=False,
        writable=True,
        path_type=Path,
    )
)
ARGUMENT_EXISTING_FILE = dict(
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path)
)


class ClickGroup(RichGroup):  # FIXME: How to set the "info_name" easier?
    def make_context(self, info_name, *args, **kwargs):
        info_name = './dev-cli.py'
        return super().make_context(info_name, *args, **kwargs)


@click.group(
    cls=ClickGroup,
    epilog=constants.CLI_EPILOG,
)
def cli():
    pass


@click.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def mypy(verbosity: int):
    """Run Mypy (configured in pyproject.toml)"""
    verbose_check_call('mypy', '.', cwd=PACKAGE_ROOT, verbose=verbosity > 0, exit_on_error=True)


cli.add_command(mypy)


@click.command()
def install():
    """
    Run pip-sync and install 'pysmartmeter' via pip as editable.
    """
    verbose_check_call('pip-sync', PACKAGE_ROOT / 'requirements.dev.txt')
    verbose_check_call('pip', 'install', '--no-deps', '-e', '.')


cli.add_command(install)


@click.command()
def safety():
    """
    Run safety check against current requirements files
    """
    verbose_check_call('safety', 'check', '-r', 'requirements.dev.txt')


cli.add_command(safety)


@click.command()
def update():
    """
    Update "requirements*.txt" dependencies files
    """
    bin_path = Path(sys.executable).parent

    verbose_check_call(bin_path / 'pip', 'install', '-U', 'pip')
    verbose_check_call(bin_path / 'pip', 'install', '-U', 'pip-tools')

    extra_env = dict(
        CUSTOM_COMPILE_COMMAND='./dev-cli.py update',
    )

    pip_compile_base = [
        bin_path / 'pip-compile',
        '--verbose',
        '--allow-unsafe',  # https://pip-tools.readthedocs.io/en/latest/#deprecations
        '--resolver=backtracking',  # https://pip-tools.readthedocs.io/en/latest/#deprecations
        '--upgrade',
        '--generate-hashes',
    ]

    # Only "prod" dependencies:
    verbose_check_call(
        *pip_compile_base,
        'pyproject.toml',
        '--output-file',
        'requirements.txt',
        extra_env=extra_env,
    )

    # dependencies + "dev"-optional-dependencies:
    verbose_check_call(
        *pip_compile_base,
        'pyproject.toml',
        '--extra=dev',
        '--output-file',
        'requirements.dev.txt',
        extra_env=extra_env,
    )

    verbose_check_call(bin_path / 'safety', 'check', '-r', 'requirements.dev.txt')

    # Install new dependencies in current .venv:
    verbose_check_call(bin_path / 'pip-sync', 'requirements.dev.txt')


cli.add_command(update)


@click.command()
def publish():
    """
    Build and upload this project to PyPi
    """
    run_unittest_cli(verbose=False, exit_after_run=False)  # Don't publish a broken state

    publish_package(
        module=pysmartmeter,
        package_path=PACKAGE_ROOT,
        distribution_name='pysmartmeter',
    )


cli.add_command(publish)


@click.command()
@click.option('--color/--no-color', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def fix_code_style(color: bool, verbosity: int):
    """
    Fix code style of all pysmartmeter source code files via darker
    """
    code_style.fix(package_root=PACKAGE_ROOT, darker_color=color, darker_verbose=verbosity > 0)


cli.add_command(fix_code_style)


@click.command()
@click.option('--color/--no-color', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def check_code_style(color: bool, verbosity: int):
    """
    Check code style by calling darker + flake8
    """
    code_style.check(package_root=PACKAGE_ROOT, darker_color=color, darker_verbose=verbosity > 0)


cli.add_command(check_code_style)


@click.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE)
def update_test_snapshot_files(verbosity: int):
    """
    Update all test snapshot files (by remove and recreate all snapshot files)
    """

    with UpdateTestSnapshotFiles(root_path=PACKAGE_ROOT, verbose=verbosity > 0):
        # Just recreate them by running tests:
        run_unittest_cli(
            extra_env=dict(
                RAISE_SNAPSHOT_ERRORS='0',  # Recreate snapshot files without error
            ),
            verbose=verbosity > 1,
            exit_after_run=False,
        )


cli.add_command(update_test_snapshot_files)


@click.command()  # Dummy command
def test():
    """
    Run unittests
    """
    run_unittest_cli()


cli.add_command(test)


@click.command()  # Dummy command
def coverage():
    """
    Run tests and show coverage report.
    """
    run_coverage()


cli.add_command(coverage)


@click.command()  # Dummy "tox" command
def tox():
    """
    Run tox
    """
    run_tox()


cli.add_command(tox)


@click.command()
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


cli.add_command(version)


######################################################################################################
# TODO:
#
# @click.command()
# @click.option('--force', **OPTION_ARGS_DEFAULT_FALSE)
# def create_default_settings(force):
#     """
#     Create a default user settings file. (Used by CI pipeline ;)
#     """
#     if not force and 'CI' not in os.environ:
#         print('We are not running in CI pipeline and "--force" not used -> Abort.')
#         sys.exit(-1)
#
#     settings_dataclass = UserSettings()
#     toml_settings = TomlSettings(
#         dir_name=SETTINGS_DIR_NAME,
#         file_name=SETTINGS_FILE_NAME,
#         settings_dataclass=settings_dataclass,
#     )
#
#     settings_path = toml_settings.file_path
#     if settings_path.is_file():
#         print(f'[green]Use settings file already exists here: {settings_path}')
#         return
#
#     document: TOMLDocument = dataclass2toml(instance=settings_dataclass)
#     doc_str = tomlkit.dumps(document, sort_keys=False)
#
#     settings_path.write_text(doc_str, encoding='UTF-8')
#     print(f'[green]Default settings file created here: {settings_path}')
#
#
# cli.add_command(create_default_settings)
#
######################################################################################################


def main():
    print_version(pysmartmeter)

    console = Console()
    rich_traceback_install(
        width=console.size.width,  # full terminal width
        show_locals=True,
        suppress=[click],
        max_frames=2,
    )

    if len(sys.argv) >= 2:
        # Check if we can just pass a command call to origin CLI:
        command = sys.argv[1]
        command_map = {
            'test': run_unittest_cli,
            'tox': run_tox,
            'coverage': run_coverage,
        }
        if real_func := command_map.get(command):
            real_func(argv=sys.argv, exit_after_run=True)

    # Execute Click CLI:
    cli()



import grp
import time
from pathlib import Path

import serial
from rich import print  # noqa
from serial.tools.list_ports_posix import comports


def get_serial(
    baudrate=9600,
    parity=serial.PARITY_EVEN,
    stopbits=serial.STOPBITS_ONE,
    bytesize=serial.SEVENBITS,
    timeout=1,
    terminator=b'\r\n',
    verbose=True,
):
    """
    Get the first "working" serial instance back.
    """
    if verbose:
        print('[cyan]Detect Serial...')

    checked_ports = []
    for port, desc, hwid in comports(include_links=False):
        checked_ports.append(port)
        if verbose:
            print('_' * 100)
            print(f'[magenta]try: {port} {desc} {hwid}')

        port_stat = Path(port).stat()
        print(f'{port} file mode:', oct(port_stat.st_mode))
        print(f'{port} user ID:', port_stat.st_uid)
        print(f'{port} user group ID:', port_stat.st_gid)
        user_group_name = grp.getgrgid(port_stat.st_gid).gr_name
        print(f'{port} user group: {user_group_name!r}')

        try:
            ser = serial.Serial(
                port,
                baudrate=baudrate,
                parity=parity,
                stopbits=stopbits,
                bytesize=bytesize,
                timeout=timeout,
            )
        except Exception as err:
            if verbose:
                print(f'[red]ERROR: {err}')
                print()
                print('Hint:')
                print(f'\t[blue]sudo usermod -a -G {user_group_name} $USER')
                print('and try again ;)')
                print('-' * 100)
            continue

        if verbose:
            print(f'Read from {ser}...', flush=True)

        for _try_count in range(5):
            try:
                data = ser.readline()
            except Exception as err:
                if verbose:
                    print(f'[yellow]ERROR: {err}')
                time.sleep(0.1)
                continue
            else:
                if verbose:
                    print(data)
                if data.endswith(terminator):
                    return ser

        if verbose:
            print('[red]Can\'t read from. Try next serial.')

    if not checked_ports:
        print('[red]No serial ports found!')


def print_detect_serial():
    ser = get_serial(verbose=True)
    if ser:
        print('[green]Found serial:')
        print(ser)
